# -*- coding: utf-8 -*-

"""
    meepo.apps.meventsourcing
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    EventSourcing app.

    This app is a simplified version of event sourcing, it only sources about
    the primary keys and their last changing time.
"""

import logging

import click

from meepo.pub import mysql_pub
from meepo.sub import es_sub

from meepo.logutils import setup_logger


@click.command()
@click.option('-m', '--master_dsn')
@click.option('-r', '--redis_dsn')
@click.option('--namespace')
@click.argument('tables', nargs=-1)
def main(master_dsn, redis_dsn, tables, namespace=None):
    setup_logger()

    logger = logging.getLogger(__name__)

    # currently only supports mysql master
    assert master_dsn and master_dsn.startswith("mysql")
    assert redis_dsn and redis_dsn.startswith("redis")

    logger.info("event sourcing tables: %s" % ", ".join(tables))
    es_sub(redis_dsn, tables, namespace)
    mysql_pub(master_dsn)



# -*- coding: utf-8 -*-

"""
    meep.apps.mprint
    ~~~~~~~~~~~~~~~~

    Event print app.
"""

import click

from meepo.pub import mysql_pub
from meepo.sub import print_sub

from meepo.logutils import setup_logger


@click.command()
@click.option("-d", "--debug", is_flag=True)
@click.option('-m', '--mysql_dsn')
@click.argument('tables', nargs=-1)
def main(mysql_dsn, tables, debug=False):
    level = "DEBUG" if debug else "INFO"
    setup_logger(level)
    print_sub(tables)
    mysql_pub(mysql_dsn, blocking=False)



# -*- coding: utf-8 -*-

import datetime
import functools
import logging
import pickle

import sqlalchemy as sa
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.exc import SQLAlchemyError

from blinker import signal

import redis
import zmq


def print_sub(tables):
    """Events print subscriber.
    """
    logger = logging.getLogger("meepo.sub.print_sub")
    logger.info("print_sub tables: %s" % ", ".join(tables))

    for table in set(tables):
        _print = lambda pk, t=table: logger.info("{} -> {}".format(t, pk))
        signal("{}_write".format(table)).connect(_print, weak=False)
        signal("{}_update".format(table)).connect(_print, weak=False)
        signal("{}_delete".format(table)).connect(_print, weak=False)


def replicate_sub(master_dsn, slave_dsn, tables=None):
    """Database replication subscriber.

    This meepo event sourcing system is based upon database primary key, so
    table should have a pk here.

    The function will subscribe to the event sourcing pk stream, retrive rows
    from master based pk and then update the slave.
    """
    logger = logging.getLogger("meepo.sub.replicate_sub")

    # sqlalchemy reflection
    logger.info("reflecting master database: {}".format(master_dsn))
    master_engine = sa.create_engine(master_dsn)
    master_base = automap_base()
    master_base.prepare(engine=master_engine, reflect=True)
    MasterSession = scoped_session(sessionmaker(bind=master_engine))

    logger.info("reflecting slave database: {}".format(slave_dsn))
    slave_engine = sa.create_engine(slave_dsn)
    slave_base = automap_base()
    slave_base.prepare(engine=slave_engine, reflect=True)
    SlaveSession = scoped_session(sessionmaker(bind=slave_engine))

    def _write_by_pk(name, pk):
        """Copy row from master to slave based on pk
        """
        MasterModel = master_base.classes[name]
        obj = MasterSession.query(MasterModel).get(pk)
        if not obj:
            logger.error("pk for {} not found in master: {}".format(name, pk))
            return

        SlaveModel = slave_base.classes[name]
        columns = [c.name for c in SlaveModel.__table__.columns]
        s_obj = SlaveModel(**{k: v
                              for k, v in obj.__dict__.items()
                              if k in columns})
        SlaveSession.add(s_obj)

        try:
            SlaveSession.commit()
        except SQLAlchemyError as e:
            SlaveSession.rollback()
            logger.exception(e)

        # cleanup
        MasterSession.close()
        SlaveSession.close()

    def _update_by_pk(name, pk):
        """Update row from master to slave based on pk
        """
        MasterModel = master_base.classes[name]
        obj = MasterSession.query(MasterModel).get(pk)

        SlaveModel = slave_base.classes[name]
        s_obj = SlaveSession.query(SlaveModel).get(pk)
        if not s_obj:
            return _write_by_pk(name, pk)

        columns = [c.name for c in SlaveModel.__table__.columns]
        for col in columns:
            try:
                val = getattr(obj, col)
            except AttributeError as e:
                continue
            setattr(s_obj, col, val)

        try:
            SlaveSession.commit()
        except SQLAlchemyError as e:
            SlaveSession.rollback()
            logger.exception(e)

        # cleanup
        MasterSession.close()
        SlaveSession.close()

    def _delete_by_pk(name, pk):
        """Copy row from slave based on pk
        """
        Model = slave_base.classes[name]
        obj = SlaveSession.query(Model).get(pk)
        if obj:
            SlaveSession.delete(obj)
        SlaveSession.commit()

        # cleanup
        SlaveSession.close()

    def _sub(table):

        def _sub_write(pk):
            logger.info("dbreplica_sub {}_write: {}".format(table, pk))
            _write_by_pk(table, pk)
        signal("%s_write" % table).connect(_sub_write, weak=False)

        def _sub_update(pk):
            logger.info("dbreplica_sub {}_update: {}".format(table, pk))
            _update_by_pk(table, pk)
        signal("%s_update" % table).connect(_sub_update, weak=False)

        def _sub_delete(pk):
            logger.info("dbreplica_sub {}_delete: {}".format(table, pk))
            _delete_by_pk(table, pk)
        signal("%s_delete" % table).connect(_sub_delete, weak=False)

    if tables:
        tables = (t for t in tables if t in slave_base.classes.keys())

    for table in tables:
        _sub(table)


def es_sub(redis_dsn, tables, namespace=None, ttl=3600*24*3):
    """EventSourcing subscriber.

    This subscriber will use redis as event sourcing storage layer.

    Note here we only needs a 'weak' event sourcing, we only record primary
    keys, which means we only care about what event happend after some time,
    and ignore how many times it happens.
    """
    logger = logging.getLogger("meepo.sub.es_sub")

    # we may accept function as namespace so we could dynamically generate it.
    # if namespace provided as string, the function return the string.
    # elif namespace not provided, generate namespace dynamically by today.
    if not callable(namespace):
        namespace = lambda: namespace if namespace else \
            "meepo:es:{}".format(datetime.date.today())

    r = redis.StrictRedis.from_url(redis_dsn)

    if ttl:
        lua_ttl = "redis.call('EXPIRE', KEYS[1], {ttl})".format(ttl=ttl)
    else:
        lua_ttl = ""

    LUA_TIME = """
    local time = redis.call('TIME')
    return tonumber(time[1])
    """
    LUA_ZADD = """
    local score = redis.call('ZSCORE', KEYS[1], ARGV[2])
    if score and ARGV[1] <= score then
        return 0
    else
        redis.call('ZADD', KEYS[1], ARGV[1], ARGV[2])
        {0}
        return 1
    end
    """.format(lua_ttl)

    r_time = r.register_script(LUA_TIME)
    r_zadd = r.register_script(LUA_ZADD)

    for table in set(tables):
        def _sub(action, pk, table=table):
            key = "%s:%s_%s" % (namespace(), table, action)
            time = r_time()
            if r_zadd(keys=[key], args=[time, str(pk)]):
                logger.info("%s_%s: %s -> %s" % (
                    table, action, pk,
                    datetime.datetime.fromtimestamp(time)))
            else:
                logger.info("%s_%s: %s -> skip" % (table, action, pk))

        signal("%s_write" % table).connect(
            functools.partial(_sub, "write"), weak=False)
        signal("%s_update" % table).connect(
            functools.partial(_sub, "update"), weak=False)
        signal("%s_delete" % table).connect(
            functools.partial(_sub, "delete"), weak=False)

    # session hooks for strict prepare-commit pattern

    def _clean_sid(sid):
        sp_all = "%s:session_prepare" % namespace()
        sp_key = "%s:session_prepare:%s" % (namespace(), sid)
        with r.pipeline() as p:
            p.srem(sp_all, sid)
            p.expire(sp_key, 60 * 60)

    def session_prepare_hook(event, sid, action):
        """Record session prepare state.
        """
        logger.info("session_prepare %s -> %s" % (action, sid))

        sp_all = "%s:session_prepare" % namespace()
        sp_key = "%s:session_prepare:%s" % (namespace(), sid)

        with r.pipeline() as p:
            p.sadd(sp_all, sid)
            p.hset(sp_key, action, pickle.dumps(event))
    signal("session_prepare").connect(session_prepare_hook, weak=False)

    def session_commit_hook(sid):
        logger.info("session_commit -> %s" % sid)
        _clean_sid(sid)
    signal("session_commit").connect(session_commit_hook, weak=False)

    def session_rollback_hook(sid):
        logger.info("session_rollback -> %s" % sid)
        _clean_sid(sid)
    signal("session_rollback").connect(session_rollback_hook, weak=False)


def zmq_sub(bind, tables, forwarder=False):
    """0mq fanout subscriber.

    This subscriber will use zeromq to publish the event to outside.
    """
    logger = logging.getLogger("meepo.sub.nano_sub")

    ctx = zmq.Context()
    pub_socket = ctx.socket(zmq.PUB)

    if forwarder:
        pub_socket.connect(bind)
    else:
        pub_socket.bind(bind)

    def _sub(table):
        for action in ("write", "update", "delete"):
            def _sub(pk, action=action):
                msg = "%s_%s %s" % (table, action, pk)
                pub_socket.send_string(msg)
                logger.debug("pub msg: %s" % msg)
            signal("%s_%s" % (table, action)).connect(_sub, weak=False)

    for table in set(tables):
        _sub(table)


def nano_sub(bind, tables):
    """Nanomsg fanout subscriber. (Experimental)

    This subscriber will use nanomsg to publish the event to outside.
    """
    logger = logging.getLogger("meepo.sub.nano_sub")

    from nanomsg import Socket, PUB

    pub_socket = Socket(PUB)
    pub_socket.bind(bind)

    def _sub(table):
        for action in ("write", "update", "delete"):
            def _sub(pk, action=action):
                msg = bytes("%s_%s %s" % (table, action, pk), 'utf-8')
                logger.debug("pub msg %s" % msg)
                pub_socket.send(msg)

            signal("%s_%s" % (table, action)).connect(_sub, weak=False)

    for table in set(tables):
        _sub(table)



# -*- coding: utf-8 -*-

"""
    meepo.apps.mnano
    ~~~~~~~~~~~~~~~~

    Nanomsg publish app.

    Sub in command line::

        $ nanocat --sub --connect tcp://127.0.0.1:6000 --format ascii

    Sub in python script::

        from nanomsg import Socket, SUB, SUB_SUBSCRIBE

        with Socket(SUB) as sub_socket:
            sub_socket.connect("tcp://127.0.0.1:6000")
            sub_socket.set_string_option(SUB, SUB_SUBSCRIBE, '')

            while True:
                print(sub_socket.recv())
"""

import logging

import click

from meepo.pub import mysql_pub
from meepo.sub import nano_sub

from meepo.logutils import setup_logger


@click.command()
@click.option('-m', '--master_dsn')
@click.option('-b', '--bind')
@click.argument('tables', nargs=-1)
def main(master_dsn, bind, tables):
    setup_logger("INFO")

    logger = logging.getLogger(__name__)

    # currently only supports mysql master
    assert master_dsn and master_dsn.startswith("mysql")
    assert bind.startswith("ipc") or bind.startswith("tcp")

    logger.info("event sourcing tables: %s" % ", ".join(tables))
    nano_sub(bind, tables)
    mysql_pub(master_dsn)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

# console_scripts
cmds = [
    "meventsourcing = meepo.apps.meventsourcing:main",
    "mnano = meepo.apps.mnano:main",
    "mprint = meepo.apps.mprint:main",
    "mreplicate = meepo.apps.mreplicate:main",
    "mzdevice = meepo.apps.mzdevice:main",
    "mzmq = meepo.apps.mzmq:main",
]


# requirements
install_requires = [
    "SQLAlchemy>=0.9.0,<1.0.0",
    "blinker>=1.3,<2.0",
    "click>=2.4,<2.5",
    "pyzmq>=14.3.1,<14.4.0",
    "redis>=2.10.1,<2.11.0",
]

mysqlbinlog_requires = [
    "mysql-replication==0.3.0",
]

# nanomsg is still in beta
nanomsg_requires = [
    "nanomsg",
]

dev_requires = [
    "flake8>=2.2.1",
    "sphinx-rtd-theme>=0.1.6",
    "sphinx>=1.2.2",
] + mysqlbinlog_requires


setup(name="meepo",
      version="0.1.3",
      description="event sourcing for databases.",
      keywords="eventsourcing event sourcing replication cache elasticsearch",
      author="Lx Yu",
      author_email="i@lxyu.net",
      packages=find_packages(exclude=['docs', 'tests']),
      entry_points={"console_scripts": cmds},
      url="https://github.com/eleme/meepo",
      license="MIT",
      zip_safe=False,
      long_description=open("README.rst").read(),
      install_requires=install_requires,
      extras_require={
          "dev": dev_requires,
          "mysqlbinlog": mysqlbinlog_requires,
          "nano": nanomsg_requires,
      },
      classifiers=[
          "Topic :: Software Development",
          "Development Status :: 3 - Alpha",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: 3.4",
      ])



# -*- coding: utf-8 -*-

"""
    meepo.apps.mzmq
    ~~~~~~~~~~~~~~~

    ZeroMQ pubsub app.

    Sub in python script::

        import zmq

        ctx = zmq.Context()
        sub_socket = ctx.socket(zmq.SUB)
        sub_socket.connect("tcp://127.0.0.1:6000")
        sub_socket.setsockopt(zmq.SUBSCRIBE, '')

        while True:
            print(sub_socket.recv_string())
"""

import logging

import click

from meepo.pub import mysql_pub
from meepo.sub import zmq_sub

from meepo.logutils import setup_logger


@click.command()
@click.option('-m', '--master_dsn')
@click.option('-b', '--bind')
@click.argument('tables', nargs=-1)
def main(master_dsn, bind, tables):
    setup_logger("INFO")

    logger = logging.getLogger(__name__)

    # currently only supports mysql master
    assert master_dsn and master_dsn.startswith("mysql")
    assert bind.startswith("ipc") or bind.startswith("tcp")

    logger.info("event sourcing tables: %s" % ", ".join(tables))
    zmq_sub(bind, tables)
    mysql_pub(master_dsn)



# -*- coding: utf-8 -*-

"""
    meepo.replicator
    ~~~~~~~~~~~~~~~~

    Meepo Replicators based on events.

    Replicate from database to targets.
"""

import logging
from multiprocessing import Process, Queue

import zmq
from zmq.utils.strtypes import asbytes


class Worker(Process):
    def __init__(self, name, queue, cb, logger_name=None):
        super(Worker, self).__init__()
        self.name = name
        self.queue = queue
        self.cb = cb

        if logger_name:
            self.logger = logging.getLogger(logger_name)

    def run(self):
        for pk in iter(self.queue.get, None):
            self.logger.info("{0} -> {1}".format(self.name, pk))
            self.cb(pk)


class ZmqReplicator(object):
    """Replicator base class.

    Trigger retrive, process, store steps to do replication.
    """

    def __init__(self, listen, name="meepo.replicator.zmq"):
        self.listen = listen
        self.workers = {}
        self.worker_queues = {}

        # replicator logger naming
        self.name = name
        self.logger = logging.getLogger(name)

        # init zmq socket
        self._ctx = zmq.Context()
        self.socket = self._ctx.socket(zmq.SUB)

    def event(self, *topics):
        """Topic callback registry.

        callback func should receive two args: topic and pk, and then process
        the replication job.
        """
        def wrapper(func):
            for topic in topics:
                self.worker_queues[topic] = Queue()
                self.workers[topic] = Worker(
                    topic, self.worker_queues[topic], func,
                    logger_name=self.name)
                self.socket.setsockopt(zmq.SUBSCRIBE, asbytes(topic))
            return func
        return wrapper

    def run(self):
        """Run replicator.

        Main process receive messages and distribute them to worker queues.
        """
        for w in self.workers.values():
            w.start()

        self.socket.connect(self.listen)
        while True:
            msg = self.socket.recv_string()
            topic, pk = msg.split()
            self.logger.debug("replicator: {0} -> {1}".format(topic, pk))
            self.worker_queues[topic].put(pk)



# -*- coding: utf-8 -*-

from logging.config import dictConfig


def setup_logger(level=None):
    dictConfig({
        'version': 1,
        'disable_existing_loggers': False,

        'root': {
            'handlers': ['console'],
            'level': 'INFO',
        },

        'loggers': {
            'meepo': {
                'handlers': ['console'],
                'propagate': False,
                'level': level or 'INFO',
            },
        },

        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'console'
            },
        },

        'formatters': {
            'console': {
                'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s',
            },
        }
    })
